# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""The Animator's rig control spaces UI."""

from functools import partial
import PySide2.QtWidgets as qtw
import PySide2.QtGui as qtg
import puppeteer.ui as ui
import puppeteer.mayax as mx
import puppeteer.rigcontrol as rigcontrol
import puppeteer.rigquery as rigquery
import puppeteer.animator as animator
from .utils import showErrorMessage


def createRigControlSpacesUI(state):
    """Create the UI for the selected rig control spaces."""
    spaceSwitches = [
        {
            'actionFn': animator.changeControlParentSpace,
            'label': 'Parent',
            'stateField': 'parentSpace',
            'placeholder': 'world',
            'icon': 'space_parent',
            'cssClass': 'control-space--parent',
        },
        {
            'actionFn': animator.changeControlPositionSpace,
            'label': 'Position',
            'stateField': 'positionSpace',
            'separator': True,
            'placeholder': 'inherit',
            'icon': 'space_position',
            'cssClass': 'control-space--position',
        },
        {
            'actionFn': animator.changeControlRotationSpace,
            'label': 'Rotation',
            'stateField': 'rotationSpace',
            'placeholder': 'inherit',
            'icon': 'space_rotation',
            'cssClass': 'control-space--rotation',
        },
        {
            'actionFn': animator.changeControlScaleSpace,
            'label': 'Scale',
            'stateField': 'scaleSpace',
            'placeholder': 'inherit',
            'icon': 'space_scale',
            'cssClass': 'control-space--scale',
        },
        {
            'actionFn': animator.changeControlAimSpace,
            'label': 'Aim',
            'stateField': 'aimSpace',
            'separator': True,
            'placeholder': 'inherit',
            'icon': 'space_aim',
            'cssClass': 'control-space--aim',
        },
    ]

    with ui.GroupVBox('Space Switching') as spacesBox:
        ui.addStyleClass(spacesBox, 'control-spaces-box')

        spacesBox.layout().setSpacing(4)
        spacesBox.layout().setContentsMargins(12, 6, 0, 6)

        for spaceSwitch in spaceSwitches:
            if 'separator' in spaceSwitch:
                spacesBox.layout().addSpacing(6)

            with ui.GroupVBox(
                spaceSwitch['label'],
                ui.loadIcon(spaceSwitch['icon']),
                collapsable=False,
            ) as box:
                ui.addStyleClass(box, f'control-space {spaceSwitch["cssClass"]}')

                with box.buttonsBar():
                    with qtw.QPushButton() as btn:
                        btn.setIcon(ui.loadIcon('replace'))
                        btn.clicked.connect(
                            partial(
                                changeSelectedControlSpace,
                                state,
                                spaceSwitch['actionFn'],
                                rigcontrol.SpaceOperation.REPLACE,
                            )
                        )

                    with qtw.QPushButton() as btn:
                        btn.setIcon(ui.loadIcon('add'))
                        btn.clicked.connect(
                            partial(
                                changeSelectedControlSpace,
                                state,
                                spaceSwitch['actionFn'],
                                rigcontrol.SpaceOperation.ADD,
                            )
                        )

                    with ui.MenuButton() as btn:
                        btn.menuAboutToShow.connect(
                            partial(
                                onSelectedControlSpaceSwitchMenu,
                                state,
                                spaceSwitch['actionFn'],
                            )
                        )

                spaces = state.selectedRigControl[spaceSwitch['stateField']]

                if spaces:
                    ui.bind(
                        box,
                        spaces,
                        partial(createControlSpaceUI, state, spaceSwitch['actionFn']),
                    )
                else:
                    with qtw.QLabel(spaceSwitch['placeholder']) as label:
                        ui.addStyleClass(label, 'control-space__empty')


def createControlSpaceUI(state: animator.AnimatorState, spaceSwitchFn, space, *_):
    """Create the UI for the control's space objects."""
    with ui.GroupVBox(collapsed=True) as box:
        box.setTitle(onSelectedControlSpaceName(state, space.target.uniqueName))
        ui.addStyleClass(box, 'control-space__target')

        with box.buttonsBar():
            with qtw.QPushButton() as btn:
                btn.setIcon(ui.loadIcon('select'))
                btn.clicked.connect(partial(selectSelectedControlSpaceTarget, state, space.target))

            with qtw.QPushButton() as btn:
                btn.setIcon(ui.loadIcon('delete'))
                btn.clicked.connect(
                    partial(
                        changeControlSpace,
                        state,
                        spaceSwitchFn,
                        state.selectedRigControl.obj,
                        space.target,
                        rigcontrol.SpaceOperation.REMOVE,
                    )
                )

        with qtw.QFormLayout() as form:
            weightField = qtw.QLineEdit()
            weightField.setValidator(qtg.QDoubleValidator())
            weightField.setText(str(round(space.weight, 3)))

            form.addRow('Weight:', weightField)

        with qtw.QPushButton('Update') as btn:
            # TODO: Use data binding instead of accesing the widget directly.
            btn.clicked.connect(
                lambda: changeControlSpace(
                    state,
                    spaceSwitchFn,
                    state.selectedRigControl.obj,
                    rigcontrol.SpaceTarget(space.target, float(weightField.text())),
                    rigcontrol.SpaceOperation.UPDATE,
                )
            )

    return box


def changeControlSpace(state: animator.AnimatorState, spaceSwitchFn, control, target, operation):
    """Change the space of a rig control."""
    spaceSwitchFn(control, target, operation, animMode=state.mode)


def changeSelectedControlSpace(state: animator.AnimatorState, spaceSwitchFn, operation):
    """Change the space for the selected rig control."""
    selectionList = mx.cmd.ls(selection=True)
    selectedControl = state.selectedRigControl.obj

    if len(selectionList) > 1:
        spaceSwitchFn(selectedControl, selectionList[1], operation, animMode=state.mode)
    else:
        showErrorMessage('No target object selected.')


def selectSelectedControlSpaceTarget(state, target):
    """Select the space target for the selected rig control."""
    if target != state.selectedRigControl.obj.world:
        target.select()


def onSelectedControlSpaceSwitchMenu(state, spaceSwitchFn, menu):
    """Fill the parent space menu with predefined spaces."""
    menu.clear()

    selectedControl = state.selectedRigControl.obj

    def addSpaceAction(menu, name, obj):
        action = menu.addAction(
            name,
            partial(
                changeControlSpace,
                state,
                spaceSwitchFn,
                selectedControl,
                obj,
                rigcontrol.SpaceOperation.REPLACE,
            ),
        )

        with action.menuItem():
            with qtw.QPushButton() as btn:
                btn.setIcon(ui.loadIcon('add'))
                btn.clicked.connect(
                    partial(
                        changeControlSpace,
                        state,
                        spaceSwitchFn,
                        selectedControl,
                        obj,
                        rigcontrol.SpaceOperation.ADD,
                    )
                )

    # add predefined spaces

    if spaceSwitchFn is animator.changeControlParentSpace:
        menu.addAction(
            'World',
            partial(
                changeControlSpace,
                state,
                spaceSwitchFn,
                selectedControl,
                None,
                rigcontrol.SpaceOperation.REPLACE,
            ),
        )
    else:
        menu.addAction(
            'Inherit',
            partial(
                changeControlSpace,
                state,
                spaceSwitchFn,
                selectedControl,
                None,
                rigcontrol.SpaceOperation.REPLACE,
            ),
        )
        menu.addAction(
            'World',
            partial(
                changeControlSpace,
                state,
                spaceSwitchFn,
                selectedControl,
                selectedControl.world,
                rigcontrol.SpaceOperation.REPLACE,
            ),
        )

    # add selected objects

    menu.addSection('Selection')

    selection = mx.cmd.ls(selection=True)
    selection.remove(selectedControl)

    if selection:
        for obj in selection:
            addSpaceAction(menu, obj.uniqueName, obj)
    else:
        menu.addAction('Empty').setDisabled(True)

    # add all controls from scene

    menu.addSection('Scene')

    allControls = rigquery.getAllControls()
    allControls.remove(selectedControl)

    if allControls:
        submenuCreator = ui.SubmenuCreator(menu)

        for control in allControls:
            submenu = submenuCreator(control.pathName)

            addSpaceAction(submenu, control.name, control)
    else:
        menu.addAction('Empty').setDisabled(True)


def onSelectedControlSpaceName(state, nodeName):
    """Change space display name."""
    try:
        controlObj = state.selectedRigControl.obj

        return 'world' if controlObj and controlObj.world == mx.Node(nodeName) else nodeName
    except mx.MayaNodeError:
        # TODO: Temporary fix, do better!
        # Error appear when deleting two controls from a 3-controls FK chain in reverse order.
        pass
