# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""The Animator's main UI."""

from functools import partial
import PySide2.QtCore as qt
import PySide2.QtWidgets as qtw
import puppeteer.ui as ui
import puppeteer.mayax as mx
import puppeteer.animator as animator
from .menubar import createMenuUI
from .toolbar import createToolbarUI
from .control_properties import createRigControlPropertiesUI
from .control_spaces import createRigControlSpacesUI
from .control_dynamics import createRigControlDynamicsUI


class MainWindow(ui.DockableWindow):
    """All the animation functionality will be accesible from this window."""

    def __init__(self):
        super().__init__('Animator', initialWidth=300, initialHeight=600)

        self.addStyleSheet('animator')

        self.state = animator.AnimatorState()

        self.createUI()

    def createUI(self):
        """Create the UI."""
        with qtw.QVBoxLayout() as mainLayout:
            self.setLayout(mainLayout)

            mainLayout.setSpacing(0)
            mainLayout.setContentsMargins(0, 0, 0, 0)

            createMenuUI()

            with qtw.QHBoxLayout():
                createToolbarUI(self.state)

                with ui.ScrollVBox() as scrollBox:
                    with qtw.QLabel('Select a Rig Control') as label:
                        ui.bindInvisibility(label, self.state.selectedRigControl, 'obj')
                        label.setAlignment(qt.Qt.AlignCenter)
                        label.setContentsMargins(0, 20, 0, 0)

                    ui.bindCreation(
                        scrollBox,
                        self.state.selectedRigControl,
                        'obj',
                        partial(createRigControlUI, self.state),
                    )

            with ui.ToggleButton() as modeBtn:
                modeBtn.addState('Setup', animator.AnimatorMode.SETUP)
                modeBtn.addState('Pose', animator.AnimatorMode.POSE)
                modeBtn.addState('Animate', animator.AnimatorMode.ANIMATE)

                ui.bind(modeBtn, self.state, 'mode')


def createRigControlUI(state):
    """Create the UI for the selected rig control."""
    with ui.GroupVBox('Selected Rig Control', collapsable=False) as mainBox:
        mainBox.layout().setContentsMargins(0, 0, 0, 0)
        mainBox.layout().setSpacing(4)

        ui.bind(mainBox.titleBarLabel(), state.selectedRigControl, 'obj', onRigControlTitleChange)
        ui.addStyleClass(mainBox, 'control-box')

        createRigControlPropertiesUI(state)
        createRigControlSpacesUI(state)
        createRigControlDynamicsUI(state)

    return mainBox


def onRigControlTitleChange(selectedControl):
    """Change Rig Contron UI title."""
    selection = mx.cmd.ls(selection=True)

    if len(selection) == 2:
        return f'{selectedControl.uniqueName} - {selection[-1].uniqueName}'

    if len(selection) > 2:
        return f'{selectedControl.uniqueName} ... {selection[-1].uniqueName}'

    return selectedControl
