# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""The Animator's toolbar UI."""

from functools import partial
import PySide2.QtCore as qt
import PySide2.QtWidgets as qtw
import puppeteer.ui as ui
import puppeteer.mayax as mx
import puppeteer.rigcontrol as rigcontrol
import puppeteer.rigquery as rigquery
import puppeteer.rigger as rigger
import puppeteer.animator as animator
import puppeteer.animutils as animutils


def createToolbarUI(animState: animator.AnimatorState):
    """Create the UI for the toolbar."""
    with qtw.QToolBar() as toolbar:
        ui.addStyleClass(toolbar, 'main-toolbar')
        toolbar.setOrientation(qt.Qt.Vertical)

        addControlActions(toolbar)

        toolbar.addSeparator()

        addControlStateActions(toolbar)

        toolbar.addSeparator()

        addPresetsActions(toolbar, animState)

        toolbar.addSeparator()

        addPoseActions(toolbar)

        toolbar.addSeparator()

        addChainActions(toolbar, animState)

        toolbar.addSeparator()

        addAnimationActions(toolbar)

        toolbar.addSeparator()

        addAnimationTransferActions(toolbar)


# --------------------------------------------------------------------------------------------------


def addControlActions(toolbar):
    """Add the control actions to the toolbar."""
    toolbar.addAction(
        ui.loadIcon('rigControl_new'),
        'Create free rig control',
        rigcontrol.createControl,
    )

    toolbar.addAction(
        ui.loadIcon('rigControl_new_withObject'),
        'Create rig control for selected objects',
        rigcontrol.createControlsForSelectedObjects,
    )

    toolbar.addAction(
        ui.loadIcon('rigControl_new_withObjectAnimated'),
        'Create rig control for selected objects while preserving the motion',
        animator.createControlsForSelectedObjects,
    )

    toolbar.addAction(
        ui.loadIcon('rigControl_deleteStatic'),
        'Delete the selected rig controls',
        lambda: rigger.deleteControls(rigcontrol.getSelectedControls()),
    )

    toolbar.addAction(
        ui.loadIcon('rigControl_deleteAnimated'),
        'Delete the selected rig controls while preserving the motion',
        lambda: animator.deleteControls(rigcontrol.getSelectedControls()),
    )


def addControlStateActions(toolbar):
    """Add the control's state actions to the toolbar."""
    for item in [
        {
            'icon': 'rigControl_show',
            'text': 'Show Rig Controls',
            'actionFn': mx.cmd.showHidden,
            'filterFn': lambda allControls: [
                control for control in allControls if not control.visibility
            ],
            'emptyText': 'No Controls Hidden',
        },
        {
            'icon': 'rigControl_hide',
            'text': 'Hide Rig Controls',
            'actionFn': mx.cmd.hide,
            'filterFn': lambda allControls: [
                control for control in allControls if control.visibility
            ],
            'emptyText': 'No Controls Visible',
        },
        {
            'icon': 'rigControl_select',
            'text': 'Select Rig Controls',
            'actionFn': mx.cmd.select,
            'filterFn': None,
            'emptyText': 'No Controls',
        },
    ]:
        action = toolbar.addAction(ui.loadIcon(item['icon']), item['text'])

        toolbar.widgetForAction(action).setPopupMode(qtw.QToolButton.InstantPopup)

        menu = qtw.QMenu()

        menu.aboutToShow.connect(
            partial(
                updateControlStateMenu,
                menu,
                item['actionFn'],
                filterFn=item['filterFn'],
                emptyText=item['emptyText'],
            )
        )

        action.setMenu(menu)


def addPresetsActions(toolbar, animState: animator.AnimatorState):
    """Add the presets actions to the toolbar."""
    action = toolbar.addAction(ui.loadIcon('rigPreset'), 'Rig Presets')

    toolbar.widgetForAction(action).setPopupMode(qtw.QToolButton.InstantPopup)

    menu = ui.Menu(toolbar)

    menu.aboutToShow.connect(partial(updatePresetsMenu, menu, animState))

    action.setMenu(menu)


def addPoseActions(toolbar):
    """Add the rest pose actions to the toolbar."""
    toolbar.addAction(
        ui.loadIcon('restPose_rig'),
        'Move rig to rest pose',
        lambda: rigger.moveControlsToRestPose(rigcontrol.getSelectedControls(), True),
    )

    toolbar.addAction(
        ui.loadIcon('restPose_selected'),
        'Move selected controls to rest pose',
        lambda: rigger.moveControlsToRestPose(rigcontrol.getSelectedControls()),
    )


def addChainActions(toolbar, animState: animator.AnimatorState):
    """Add the chain actions to the toolbar."""
    toolbar.addAction(
        ui.loadIcon('chain_fk'),
        'FK chain',
        lambda: animator.switchControlsToFK(
            rigcontrol.getSelectedControls(),
            animMode=animState.mode,
        ),
    )

    toolbar.addAction(
        ui.loadIcon('chain_rfk'),
        'Reverse FK chain',
        lambda: animator.switchControlsToReversedFK(
            rigcontrol.getSelectedControls(),
            animMode=animState.mode,
        ),
    )

    toolbar.addAction(
        ui.loadIcon('chain_ik'),
        'IK chain',
        lambda: animator.switchControlsToIK(
            rigcontrol.getSelectedControls(),
            animMode=animState.mode,
        ),
    )


def addAnimationActions(toolbar):
    """Add the animation actions to the toolbar."""
    toolbar.addAction(
        ui.loadIcon('anim_setKey_rig'),
        'Set Key on Entire Rig',
        animator.keySelectedRig,
    )

    toolbar.addAction(
        ui.loadIcon('anim_bake_rig'),
        'Bake Rig',
        lambda: animator.bakeSelectedRig(timeRange=animutils.getSelectedTimeRange()),
    )

    toolbar.addAction(
        ui.loadIcon('anim_setKey_selected'),
        'Set Key on Selected Objects',
        animutils.keySelectedObjects,
    )

    toolbar.addAction(
        ui.loadIcon('anim_bake_selected'),
        'Bake Selected Objects',
        lambda: animutils.bakeSelectedObjects(timeRange=animutils.getSelectedTimeRange()),
    )


def addAnimationTransferActions(toolbar):
    """Add the animation transfer actions to the toolbar."""
    toolbar.addAction(
        ui.loadIcon('anim_transfer'),
        'Transfer animation',
        animator.transferAnimationBetweenSelectedObjects,
    )

    toolbar.addAction(
        ui.loadIcon('anim_transfer_position'),
        'Transfer position animation',
        animator.transferPositionAnimationBetweenSelectedObjects,
    )

    toolbar.addAction(
        ui.loadIcon('anim_transfer_rotation'),
        'Transfer rotation animation',
        animator.transferRotationAnimationBetweenSelectedObjects,
    )

    toolbar.addAction(
        ui.loadIcon('anim_transfer_scale'),
        'Transfer scale animation',
        animator.transferScaleAnimationBetweenSelectedObjects,
    )


# --------------------------------------------------------------------------------------------------


def updateControlStateMenu(menu, actionFn, filterFn, emptyText):
    """Update the control state menu."""
    menu.clear()

    def onSubmenuCreation(submenu: qtw.QMenu, path: str):
        subControls = [
            control for control in allControls if control.pathName.startswith(f'{path}|')
        ]

        submenu.addAction('all', partial(actionFn, subControls))
        submenu.addSeparator()

    allControls = rigquery.getAllControls()

    if filterFn:
        allControls = filterFn(allControls)

    if not allControls:
        menu.addAction(emptyText).setDisabled(True)

        return

    menu.addAction('all', partial(actionFn, allControls))
    menu.addSeparator()

    submenuCreator = ui.SubmenuCreator(menu)

    for control in allControls:
        submenu = submenuCreator(control.pathName, onSubmenuCreation)

        submenu.addAction(control.name, partial(actionFn, control))


def updatePresetsMenu(menu, animState: animator.AnimatorState):
    """Update the rig presets menu."""
    menu.clear()

    rigs = []

    def applyPreset(rigRoot, preset):
        animator.applyPreset(rigRoot, preset, animMode=animState.mode)

    for control in rigcontrol.getSelectedControls():
        rigRoot = rigquery.getRigRootFromControl(control)

        if rigRoot and rigRoot not in rigs:
            rigs.append(rigRoot)

        # TODO: Cover root selection.

    if not rigs:
        menu.addAction('No Rig Selected').setDisabled(True)
        return

    if len(rigs) > 1:
        menu.addAction('Too Many Rigs Selected').setDisabled(True)
        return

    rigRoot = rigs[0]
    presets = rigquery.getRigPresets(rigRoot, ordered=True)

    if presets:
        submenuCreator = ui.SubmenuCreator(menu)

        for preset in presets:
            presetName = preset.name.split('|')[-1]
            submenu = submenuCreator(preset.name)

            action = submenu.addAction(presetName, partial(applyPreset, rigRoot, preset))

            with action.menuItem():
                with qtw.QPushButton() as btn:
                    btn.setIcon(ui.loadIcon('delete'))
                    btn.clicked.connect(partial(rigger.deletePreset, rigRoot, preset.name))
    else:
        menu.addAction('No Rig Presets').setDisabled(True)

    menu.addSeparator()

    menu.addAction('Save Preset', partial(savePreset, rigRoot))


# --------------------------------------------------------------------------------------------------


def savePreset(rigRoot):
    """Save the rig preset."""
    name, submitted = qtw.QInputDialog.getText(
        None,
        'Save Rig Preset',
        '<b>Enter a preset name:</b><br>(use | to nest: category|...|name)',
    )

    if not submitted:
        return

    name = rigger.sanitizePresetName(name)

    if not name:
        msgBox = qtw.QMessageBox(qtw.QMessageBox.Warning, 'Save Rig Preset', 'Invalid preset name.')
        msgBox.exec()

        savePreset(rigRoot)

        return

    rigger.savePreset(rigRoot, name)
