# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""Animation utility functions."""

from __future__ import annotations
import contextlib
import maya.mel as mel
import puppeteer.mayax as mx
import puppeteer.rigcontrol


@mx.undoable
def bake(objects, timeRange=None, deleteStaticCurves=False):
    """Bake simulation."""
    if not timeRange:
        timeRange = (
            mx.cmd.playbackOptions(query=True, animationStartTime=True),
            mx.cmd.playbackOptions(query=True, animationEndTime=True),
        )

    mx.cmd.refresh(suspend=True)

    mx.cmd.bakeResults(
        objects,
        simulation=True,
        time=timeRange,
        sampleBy=1,
        oversamplingRate=1,
        disableImplicitControl=True,
        preserveOutsideKeys=True,
        minimizeRotation=True,
    )

    mx.cmd.filterCurve(objects, filter='euler')

    if deleteStaticCurves:
        mx.cmd.delete(objects, staticChannels=True)

    mx.cmd.refresh(suspend=False)

    # TODO: Find a better way.
    # In a parent-child space (other spaces too),
    # when moving the parent while preserving the animation,
    # the control target doesn't get updated without marking it dirty.
    mx.cmd.evalDeferred('maya.cmds.dgdirty(allPlugs=True)')


@mx.undoable
def bakeSelectedObjects(timeRange=None):
    """Bake simulation for the selected objects."""
    selectedObjects = mx.cmd.ls(selection=True)

    if selectedObjects:
        bake(selectedObjects, timeRange=timeRange)


@mx.undoable
def transferAnimation(fromObj, toObj):
    """Transfer animation between two objects."""
    mx.cmd.cutKey(toObj, clear=True, time=(), attribute=['translate', 'rotate', 'scale'])

    parentConstraint = mx.cmd.parentConstraint(fromObj, toObj, maintainOffset=True)[0]
    scaleConstraint = mx.cmd.scaleConstraint(fromObj, toObj, maintainOffset=True)[0]

    tempNodes = [parentConstraint, scaleConstraint]

    if puppeteer.rigcontrol.isControl(toObj):
        # TODO: Remove if the rig control is no longer using the joint for its transform node.
        scaleConstraint['constraintScaleCompensate'].disconnectInput()
        scaleConstraint.constraintScaleCompensate = False

        scaleOffsetNode = mx.cmd.createNode('multiplyDivide', skipSelect=True)
        scaleConstraint['constraintScale'].connect(scaleOffsetNode['input1'])
        toObj['inverseScale'].connect(scaleOffsetNode['input2'])

        scaleOffsetNode['outputX'].connect(toObj['scaleX'], force=True)
        scaleOffsetNode['outputY'].connect(toObj['scaleY'], force=True)
        scaleOffsetNode['outputZ'].connect(toObj['scaleZ'], force=True)

        tempNodes.append(scaleOffsetNode)

    bake(toObj)

    mx.cmd.delete(tempNodes)


@mx.undoable
def transferPositionAnimation(fromObj, toObj):
    """Transfer position animation between two objects."""
    mx.cmd.cutKey(toObj, clear=True, time=(), attribute=['translate'])

    constraint = mx.cmd.pointConstraint(fromObj, toObj, maintainOffset=True)[0]

    bake(toObj)

    constraint.delete()


@mx.undoable
def transferRotationAnimation(fromObj, toObj):
    """Transfer rotation animation between two objects."""
    mx.cmd.cutKey(toObj, clear=True, time=(), attribute=['rotate'])

    constraint = mx.cmd.orientConstraint(fromObj, toObj, maintainOffset=True)[0]

    bake(toObj)

    constraint.delete()


@mx.undoable
def transferScaleAnimation(fromObj, toObj):
    """Transfer scale animation between two objects."""
    mx.cmd.cutKey(toObj, clear=True, time=(), attribute=['scale'])

    constraint = mx.cmd.scaleConstraint(fromObj, toObj, maintainOffset=True)[0]

    tempNodes = [constraint]

    if puppeteer.rigcontrol.isControl(toObj):
        # TODO: Remove if the rig control is no longer using the joint for its transform node.
        constraint['constraintScaleCompensate'].disconnectInput()
        constraint.constraintScaleCompensate = False

        scaleOffsetNode = mx.cmd.createNode('multiplyDivide', skipSelect=True)
        constraint['constraintScale'].connect(scaleOffsetNode['input1'])
        toObj['inverseScale'].connect(scaleOffsetNode['input2'])

        scaleOffsetNode['outputX'].connect(toObj['scaleX'], force=True)
        scaleOffsetNode['outputY'].connect(toObj['scaleY'], force=True)
        scaleOffsetNode['outputZ'].connect(toObj['scaleZ'], force=True)

        tempNodes.append(scaleOffsetNode)

    bake(toObj)

    mx.cmd.delete(tempNodes)


@mx.undoable
def transferAnimationToLocators(objects, timeRange=None):
    """Transfer animation from objects to locators."""
    bakedLocators = []
    bakedConstraints = []

    if not objects:
        return bakedLocators

    for obj in objects:
        locator = mx.cmd.spaceLocator()[0]
        locator.parent = obj.parent
        locator.name = f'{obj.name}_baked'
        locator.inheritsTransform = False

        parentConstraint = mx.cmd.parentConstraint(obj, locator)[0]
        scaleConstraint = mx.cmd.scaleConstraint(obj, locator)[0]

        if puppeteer.rigcontrol.isControl(obj):
            scaleOffsetNode = mx.cmd.createNode('multiplyDivide', skipSelect=True)
            scaleOffsetNode.operation = 2  # divide

            obj['scale'].connect(scaleOffsetNode['input1'])
            obj['inverseScale'].connect(scaleOffsetNode['input2'])

            scaleOffsetNode['output'].connect(scaleConstraint['target[0].targetScale'], force=True)

            bakedConstraints.append(scaleOffsetNode)

        bakedLocators.append(locator)
        bakedConstraints.append(parentConstraint)
        bakedConstraints.append(scaleConstraint)

    bake(bakedLocators, timeRange)

    mx.cmd.delete(bakedConstraints)

    return bakedLocators


@mx.undoable
def transferAnimationFromLocators(locators, objects, timeRange=None):
    """Transfer animation from locators to objects."""
    if not locators:
        return

    bakedConstraints = []

    mx.cmd.cutKey(objects, clear=True, time=(), attribute=['translate', 'rotate', 'scale'])

    for index, obj in enumerate(objects):
        parentConstraint = mx.cmd.parentConstraint(locators[index], obj)[0]
        scaleConstraint = mx.cmd.scaleConstraint(locators[index], obj)[0]

        if puppeteer.rigcontrol.isControl(obj):
            # TODO: Remove if the rig control is no longer using the joint for its transform node.
            scaleConstraint['constraintScaleCompensate'].disconnectInput()
            scaleConstraint.constraintScaleCompensate = False

            scaleOffsetNode = mx.cmd.createNode('multiplyDivide', skipSelect=True)
            scaleConstraint['constraintScale'].connect(scaleOffsetNode['input1'])
            obj['inverseScale'].connect(scaleOffsetNode['input2'])

            scaleOffsetNode['outputX'].connect(obj['scaleX'], force=True)
            scaleOffsetNode['outputY'].connect(obj['scaleY'], force=True)
            scaleOffsetNode['outputZ'].connect(obj['scaleZ'], force=True)

            bakedConstraints.append(scaleOffsetNode)

        bakedConstraints.append(parentConstraint)
        bakedConstraints.append(scaleConstraint)

    bake(objects, timeRange)

    mx.cmd.delete(bakedConstraints)
    mx.cmd.delete(locators)


@contextlib.contextmanager
def preserveMotion(objects, preserveKeys=False):
    """Preserve the objects' motion after changing their transforms."""
    undoName = 'preserveMotion'

    mx.cmd.undoInfo(openChunk=True, chunkName=undoName)

    selection = mx.cmd.ls(selection=True)
    bakedLocators = transferAnimationToLocators(objects)

    # TODO: Shouldn't be necessary if `transferAnimationToLocators` wouldn't select various nodes.
    mx.cmd.select(selection)

    yield bakedLocators

    newObjects = []
    newBakedLocators = []

    for i, obj in enumerate(objects):
        if obj.apiObjectHandle.isValid():
            newObjects.append(obj)
            newBakedLocators.append(bakedLocators[i])
        else:
            try:
                bakedLocators[i].delete()
            except mx.MayaNodeError:
                continue

    objects = newObjects
    bakedLocators = newBakedLocators

    if preserveKeys:
        oldKeys = [getFramesWithTransformKeys(obj) for obj in objects]

    selection = mx.cmd.ls(selection=True)
    transferAnimationFromLocators(bakedLocators, objects)

    if preserveKeys:
        for i, obj in enumerate(objects):
            if oldKeys[i]:
                keysToRemove = [
                    (key,) for key in getFramesWithTransformKeys(obj) if key not in oldKeys[i]
                ]
                mx.cmd.cutKey(obj, time=keysToRemove, clear=True)

    # TODO: Shouldn't be necessary if `transferAnimationToLocators` wouldn't select various nodes.
    mx.cmd.select(selection)

    mx.cmd.undoInfo(closeChunk=True, chunkName=undoName)


@mx.undoable
def keySelectedObjects():
    """Key selected objects."""
    selection = mx.cmd.ls(selection=True)

    if selection:
        mx.cmd.setKeyframe(selection)


def getFramesWithTransformKeys(obj):
    """Get all the frames that have transform (translate/rotate/scale) keys."""
    frames = []

    for curveName in ('translate', 'rotate', 'scale'):
        for frame in mx.cmd.keyframe(obj[curveName], query=True) or []:
            if frame not in frames:
                frames.append(frame)

    frames.sort()

    return frames


def getSelectedTimeRange() -> tuple[float] | None:
    """Get the selected time range from the global time control."""
    timeControlName = mel.eval('$tmpVar=$gPlayBackSlider')

    if mx.cmd.timeControl(timeControlName, query=True, rangeVisible=True):
        return tuple(mx.cmd.timeControl(timeControlName, query=True, rangeArray=True))

    return None
