# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""Watch for code changes and reload the python modules."""

import os
import PySide2.QtCore as qt
import PySide2.QtWidgets as qtw
import maya.api.OpenMaya as om
import maya.cmds as cmds


# --------------------------------------------------------------------------------------------------


fsWatcher = None  # pylint: disable=invalid-name
hotReloadCallbackId = None  # pylint: disable=invalid-name
hotReloadRequired = False  # pylint: disable=invalid-name


def enable():
    """Enable hot reload."""
    from puppeteer import log

    global fsWatcher  # pylint: disable=invalid-name,global-statement

    fsWatcher = qt.QFileSystemWatcher(getPathsToWatch())
    fsWatcher.directoryChanged.connect(onDirectoryChanged)
    fsWatcher.fileChanged.connect(onFileChanged)

    qtw.QApplication.instance().applicationStateChanged.connect(onApplicationStateChanged)

    log.info('Hot Reload enabled.')


def disable():
    """Disable hot reload."""
    from puppeteer import log

    try:
        registeredPaths = fsWatcher.directories() + fsWatcher.files()
        fsWatcher.removePaths(registeredPaths)

        qtw.QApplication.instance().applicationStateChanged.disconnect(onApplicationStateChanged)

        log.info('Hot Reload disabled.')
    except AttributeError:
        log.warning('Hot Reload could not be disabled.')


def hotReload():
    """Do hot reload."""
    from puppeteer import ui
    from puppeteer import log
    from puppeteer import settings
    from puppeteer.utils import removePyModules

    windows = ui.findWidgetsByClass(ui.Window)
    windowsToOpen = []

    for window in windows:
        cls = window.__class__
        windowsToOpen.append((cls.__module__, cls.__name__))

        window.close()

    removePyModules('puppeteer', ('puppeteer', 'puppeteer.hotreload'))

    extraPackage = settings.getValue('hotReloadExtraPackage')
    if extraPackage:
        removePyModules(extraPackage, (extraPackage,))

    for module, cls in windowsToOpen:
        script = f'from {module} import {cls};{cls}().show()'
        cmds.evalDeferred(script)

    log.info('Hot Reload complete.')


def hotReloadDelayed():
    """Do hot reload after a short delay."""
    global hotReloadCallbackId  # pylint: disable=invalid-name,global-statement

    def doReload(*_):
        global hotReloadCallbackId  # pylint: disable=invalid-name,global-statement

        om.MMessage.removeCallback(hotReloadCallbackId)
        hotReloadCallbackId = None

        hotReload()

    if hotReloadCallbackId:
        om.MMessage.removeCallback(hotReloadCallbackId)

    hotReloadCallbackId = om.MTimerMessage.addTimerCallback(0.25, doReload)


def getPathsToWatch():
    """Get the file/directory paths to watch for changes."""
    from puppeteer import settings
    from puppeteer.utils import findFiles

    dirPath = os.path.dirname(os.path.realpath(__file__))
    paths = [dirPath] + findFiles(dirPath, nameFilters=['*.py'])

    extraPath = settings.getValue('hotReloadExtraPath')
    if extraPath:
        paths += [extraPath] + findFiles(extraPath, nameFilters=['*.py', '*.qss'])

    qssDirPath = os.path.realpath(os.path.join(dirPath, '..', 'qss'))
    paths += [qssDirPath] + findFiles(qssDirPath, nameFilters=['*.qss'])

    return paths


def onDirectoryChanged():
    """Update the watched files."""
    registeredPaths = fsWatcher.directories() + fsWatcher.files()
    fsWatcher.removePaths(registeredPaths)
    fsWatcher.addPaths(getPathsToWatch())


def onFileChanged(filePath):
    """Request hot reload."""
    global hotReloadRequired  # pylint: disable=invalid-name,global-statement

    if filePath.endswith('.qss'):
        hotReloadDelayed()
    else:
        hotReloadRequired = True


def onApplicationStateChanged(state):
    """Check if a Hot Reload is needed on application activation."""
    global hotReloadRequired  # pylint: disable=invalid-name,global-statement

    if state == qt.Qt.ApplicationActive and hotReloadRequired:
        hotReload()

        hotReloadRequired = False
