# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""Rig control's links functionality."""

from enum import IntEnum
import puppeteer.mayax as mx


class LinkStyle(IntEnum):
    """Rig control's link style."""

    SOLID = 1
    DASHED = 2
    DOTTED = 3


class LinkDirection(IntEnum):
    """Rig control's link direction."""

    NONE = 0
    FORWARD = 1
    BACKWARD = 2


class LinkTarget:
    """Rig control's link target."""

    def __init__(
        self,
        target: mx.Node,
        color: tuple = (1.0, 1.0, 1.0),
        style: LinkStyle = LinkStyle.SOLID,
        direction: LinkDirection = LinkDirection.NONE,
    ):
        self.target = target
        self.color = color
        self.style = style
        self.direction = direction

    def __eq__(self, other):
        return self.target == other

    def __ne__(self, other):
        return self.target != other


@mx.undoable
def linkControl(control, linkTarget):
    """Link the rig control to an object."""
    if not isinstance(linkTarget, LinkTarget):
        linkTarget = LinkTarget(linkTarget)

    if linkTarget in getControlLinks(control):
        return

    controlShape = control.shapes[0]

    if controlShape['links[0].linkTarget'].input(shapes=True) != controlShape:
        # add itself to trigger a viewport redraw when moved
        controlShape['worldMatrix'].connect(controlShape['links[0].linkTarget'])

    linksIndices = mx.cmd.getAttr(controlShape['links'], multiIndices=True)

    # find an empty link (undo or target deletion causes that)
    for linkIndex in linksIndices[1:]:  # skip itself
        if controlShape[f'links[{linkIndex}].linkTarget'].input() is None:
            break
    else:
        linkIndex = linksIndices[-1] + 1

    linkTarget.target['worldMatrix'].connect(controlShape[f'links[{linkIndex}].linkTarget'])

    controlShape[f'links[{linkIndex}].linkColor'].value = linkTarget.color
    controlShape[f'links[{linkIndex}].linkStyle'].value = linkTarget.style
    controlShape[f'links[{linkIndex}].linkDirection'].value = linkTarget.direction


@mx.undoable
def unlinkControl(control, linkTarget=None):
    """Unlink the rig control from an object.

    If `linkTarget` is `None` all links will be removed.
    """
    controlShape = control.shapes[0]

    linksIndices = mx.cmd.getAttr(controlShape['links'], multiIndices=True) or []

    for linkIndex in linksIndices:
        if not linkTarget:
            mx.cmd.removeMultiInstance(controlShape[f'links[{linkIndex}]'], b=True)
        elif linkTarget == controlShape[f'links[{linkIndex}].linkTarget'].input():
            mx.cmd.removeMultiInstance(controlShape[f'links[{linkIndex}]'], b=True)

            if len(linksIndices) == 2:
                # remove itself
                mx.cmd.removeMultiInstance(controlShape['links[0]'], b=True)

            return


def getControlLinks(control):
    """Get control links."""
    controlShape = control.shapes[0]
    linksIndices = mx.cmd.getAttr(controlShape['links'], multiIndices=True) or []

    links = []

    for linkIndex in linksIndices[1:]:  # ignore itself
        links.append(
            LinkTarget(
                target=controlShape[f'links[{linkIndex}].linkTarget'].input(),
                color=controlShape[f'links[{linkIndex}].linkColor'].value,
                style=LinkStyle(controlShape[f'links[{linkIndex}].linkStyle'].value),
                direction=LinkDirection(controlShape[f'links[{linkIndex}].linkDirection'].value),
            )
        )

    return links
