# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""Rig control's main functionality."""

import puppeteer.mayax as mx
import puppeteer.mtxconstrain as mtxconstrain
import puppeteer.rigutils as rigutils


@mx.undoable
def createControl(target=None, name=None, parent=None, skipSelect=False):
    """Create a rig control."""
    if name:
        ctrlName = name
    elif target:
        ctrlName = f'{target.name}_ctrl'
    else:
        ctrlName = 'free_ctrl'

    control = mx.cmd.createNode('joint', name=ctrlName, parent=parent, skipSelect=True)
    control.drawStyle = 2  # None
    control.inheritsTransform = False
    control['inheritsTransform'].locked = True
    control['visibility'].keyable = False
    control['radius'].channelBox = False

    controlShape = mx.cmd.createNode(
        'puppeteerRigControl',
        name=f'{ctrlName}Shape' + ('#' if control.name != ctrlName else ''),
        parent=control,
        skipSelect=True,
    )

    controlShape['rotateBuffer'].connect(control['jointOrient'])
    controlShape['scaleBuffer'].connect(control['inverseScale'])
    controlShape['parentOffset'].connect(control['offsetParentMatrix'])

    ctrlWorld = mx.cmd.createNode('transform', name='world', parent=control, skipSelect=True)
    ctrlWorld.inheritsTransform = False
    ctrlWorld.hiddenInOutliner = True

    control.addAttr('world', ctrlWorld, type=mx.Node)
    control.addAttr('target', type=mx.Node)
    control.addAttr('constraint', type=mx.Node)

    if target:
        control.rotateOrder = target.rotateOrder

        mx.cmd.matchTransform(control, target)
        constraint = mtxconstrain.parentConstraint(control, target)

        control.target = target
        control.constraint = constraint

    control.addAttr('restPosition', control.worldPosition)
    control.addAttr('restRotation', control.worldRotation)
    control.addAttr('restScale', control.worldScale)

    if not skipSelect:
        control.select()

    return control


@mx.undoable
def createControlsForSelectedObjects():
    """Create rig controls for the selected objects."""
    selectionList = mx.cmd.ls(selection=True)
    controls = []

    for obj in selectionList:
        controls.append(createControl(obj, skipSelect=True))

    mx.cmd.select(controls)

    return controls


@mx.undoable
def createPoleVectorControl(ikHandle, name=None, parent=None, skipSelect=True):
    """Create a rig control for the pole vector of an IK handle."""
    controlName = name or f'{ikHandle.startJoint.children[0].name}_ikPole'

    control = createControl(name=controlName, parent=parent, skipSelect=skipSelect)

    rigutils.placeObjectToPoleVector(control, ikHandle)

    control.target = ikHandle
    control.constraint = mx.cmd.poleVectorConstraint(control, ikHandle)[0]

    control.restPosition = control.worldPosition
    control.restRotation = control.worldRotation
    control.restScale = control.worldScale

    return control


def isControl(obj):
    """Check if the provided object is a rig control."""
    try:
        return obj.shapes[0].type == 'puppeteerRigControl'
    except (IndexError, AttributeError):
        return False


def getControlFromTarget(target):
    """Get the control from its target."""
    connectedAttributes = target['message'].outputs(type='transform', plugs=True)

    for attr in connectedAttributes:
        if attr.name == 'target' and isControl(attr.node):
            return attr.node

    return None


def getSelectedControls():
    """Get the selected controls."""
    return [obj for obj in mx.cmd.ls(selection=True) if isControl(obj)]


# --------------------------------------------------------------------------------------------------


@mx.undoable
def retargetControl(control, newTarget, maintainOffset=True):
    """Change a rig control's target."""
    control.constraint.delete()

    if not maintainOffset:
        mx.cmd.matchTransform(control, newTarget, position=True, rotation=True)
        control.resetTransform()

    control.constraint = mtxconstrain.parentConstraint(control, newTarget)
    control.target = newTarget


@mx.undoable
def freeControl(control):
    """Remove the relationship of a rig control with its target."""
    control.constraint.delete()
    control.target = None


@mx.undoable
def deleteControl(control):
    """Delete a rig control."""
    # TODO: Fix errors when undoing the deletion.
    mtxconstrain.deleteParentConstraints(control)

    control.delete()


@mx.undoable
def disconnectControl(control, removeConstraint=False):
    """Disconnect a rig control from its target."""
    if removeConstraint:
        control.constraint.delete()
    else:
        control.constraint['matrixSum'].disconnect(control.target['offsetParentMatrix'])


@mx.undoable
def reconnectControl(control, updateConstraint=False):
    """Reconnect a rig control to its target."""
    constraint = control.constraint

    if constraint and not updateConstraint:
        control.constraint['matrixSum'].connect(control.target['offsetParentMatrix'])
    else:
        if constraint:
            constraint.delete()

        control.constraint = mtxconstrain.parentConstraint(control, control.target)
