# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""Rig data query functionality."""

from __future__ import annotations
import collections
import puppeteer.mayax as mx
import puppeteer.rigcontrol as rigcontrol
import puppeteer.rigdefinition as rigdef


IKChain = collections.namedtuple('IKChain', ['controls', 'poleControl', 'joints'])


def getRigRootFromControl(control):
    """Get the rig root from a control."""
    rigRoot = None
    parent = control.parent

    while parent:
        if not parent.shapes:
            rigRoot = parent

        parent = parent.parent

    return rigRoot


def getAllControls(rigRoot=None):
    """Get all controls under a rig root, or from scene."""
    if not rigRoot:
        return [shape.parent for shape in mx.cmd.ls(type='puppeteerRigControl', dagObjects=True)]

    controls = (
        mx.cmd.listRelatives(
            rigRoot,
            allDescendents=True,
            type='puppeteerRigControl',
            fullPath=True,
        )
        or []
    )

    return [shape.parent for shape in controls]


# --------------------------------------------------------------------------------------------------


def getRigPresets(rigRoot, ordered=False) -> list[rigdef.RigDefinition.Preset]:
    """Get the rig presets."""
    presets = []

    if not rigRoot.hasAttr('presets'):
        return presets

    presets = rigdef.RigDefinition().loadPresets(rigRoot.presets).presets

    if ordered:
        presets.sort(key=lambda preset: preset.name)

    return presets


# --------------------------------------------------------------------------------------------------


def getFKChainFromSingleControl(control):
    """Get from a single control all the controls that are part of a FK chain."""
    chain = [control]

    while True:
        spaces = rigcontrol.getControlParentSpace(chain[0])

        if len(spaces) != 1 or not rigcontrol.isControl(spaces[0].target):
            break

        chain.insert(0, spaces[0].target)

    while True:
        drivenControls = rigcontrol.getControlsDrivenByParentSpace(chain[-1])

        if len(drivenControls) != 1:
            break

        chain.append(drivenControls[0])

    return chain


def getIKChainFromSingleControl(control):
    """Get from a single control all the controls that are part of an IK chain."""
    chain = None
    jointsChain = []

    for attr in control['message'].outputs(type='joint', plugs=True):
        if attr.name == 'ikControl' or attr.name == 'ikPoleControl':
            ikJoint = attr.node

            jointParent = ikJoint.parent
            while jointParent and jointParent.type == 'joint':
                jointsChain.insert(0, jointParent)
                jointParent = jointParent.parent

            jointsChain.append(ikJoint)

            joints = (
                mx.cmd.listRelatives(
                    ikJoint,
                    allDescendents=True,
                    type='joint',
                    fullPath=True,
                )
                or []
            )

            jointsChain.extend(reversed(joints))

            break

    if jointsChain:
        chain = IKChain(
            controls=[joint.ikControl for joint in jointsChain],
            poleControl=jointsChain[-1].ikPoleControl,
            joints=jointsChain,
        )

    return chain
