# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""Utility functionality for rigging."""

import math
import maya.api.OpenMaya as om
import puppeteer.mayax as mx


@mx.undoable
def placeObjectToPoleVector(obj, ikHandle):
    """Set the position and rotation of an object to match the pole vector of an IK handle."""
    chain = [
        ikHandle.startJoint,
        ikHandle.startJoint.children[0],
        ikHandle.endEffector,
    ]

    position, rotation = calculatePoleVectorPositionAndRotation(chain)

    obj.worldPosition = position
    obj.worldRotation = rotation


def calculatePoleVectorPositionAndRotation(chain, objSpace=None, time=None):
    """Calculate the position and rotation of an object to match the pole vector of an IK handle."""
    if isinstance(chain[0], mx.Vector):
        start = chain[0]
        middle = chain[1]
        end = chain[-1]
    elif time is not None:
        start = chain[0].worldPositionAt(time)
        middle = chain[1].worldPositionAt(time)
        end = chain[-1].worldPositionAt(time)
    else:
        start = chain[0].worldPosition
        middle = chain[1].worldPosition
        end = chain[-1].worldPosition

    startEnd = end - start
    startMid = middle - start
    midEnd = end - middle

    try:
        projection = start + ((startMid * startEnd) / (startEnd * startEnd)) * startEnd
    except ZeroDivisionError:
        # TODO: Is this the right thing to do?
        projection = mx.Vector.kZeroVector

    direction = (middle - projection).normal()
    distance = (startMid.length() + midEnd.length()) * 0.5

    position = middle + direction * distance

    rotationAxes = []
    rotationAxes.append((startMid ^ startEnd).normal())  # x
    rotationAxes.append((direction ^ rotationAxes[0]).normal())  # y
    rotationAxes.append(direction)  # z

    transformationMatrix = om.MTransformationMatrix(
        om.MMatrix(
            (
                (rotationAxes[0].x, rotationAxes[0].y, rotationAxes[0].z, 0),
                (rotationAxes[1].x, rotationAxes[1].y, rotationAxes[1].z, 0),
                (rotationAxes[2].x, rotationAxes[2].y, rotationAxes[2].z, 0),
                (position.x, position.y, position.z, 1),
            )
        )
    )

    if objSpace:
        transformationMatrix = om.MTransformationMatrix(
            transformationMatrix.asMatrix() * objSpace.parentInverseMatrix
        )
        position = transformationMatrix.translation(om.MSpace.kTransform)

    rotation = transformationMatrix.rotation()
    rotation = mx.Vector(
        math.degrees(rotation.x),
        math.degrees(rotation.y),
        math.degrees(rotation.z),
    )

    return (position, rotation)


def calculateAimConstraintVectors(obj, target):
    """Calculate the aim constraint vectors using the target for object rotation up."""
    objWorldMatrix = obj.worldMatrix
    targetWorldMatrix = target.worldMatrix
    aimDirection = (target.worldPosition - obj.worldPosition).normal()

    worldAxes = [
        mx.Vector(1, 0, 0),
        mx.Vector(0, 1, 0),
        mx.Vector(0, 0, 1),
        mx.Vector(-1, 0, 0),
        mx.Vector(0, -1, 0),
        mx.Vector(0, 0, -1),
    ]

    # Convert the objects' local axes to world space.
    objWorldAxes = [worldAxis * objWorldMatrix for worldAxis in worldAxes]
    targetWorldAxes = [worldAxis * targetWorldMatrix for worldAxis in worldAxes]

    # Find the aim axis by checking which one points closer to the target.
    aimAxesRatios = [objAxis * aimDirection for objAxis in objWorldAxes]
    aimAxisIndex = aimAxesRatios.index(max(aimAxesRatios))
    aimAxis = worldAxes[aimAxisIndex]

    # Remove aim axis from those available to properly find the up axis.
    del worldAxes[aimAxisIndex]
    #
    oppositeAimAxisIndex = worldAxes.index(aimAxis * -1)
    del worldAxes[oppositeAimAxisIndex]
    #
    del objWorldAxes[aimAxisIndex]
    del objWorldAxes[oppositeAimAxisIndex]
    del targetWorldAxes[aimAxisIndex]
    del targetWorldAxes[oppositeAimAxisIndex]

    # Find the up axis by checking which one matches the target's axis better.
    upAxesRatios = [objAxis * targetWorldAxes[i] for i, objAxis in enumerate(objWorldAxes)]
    upAxis = worldAxes[upAxesRatios.index(max(upAxesRatios))]

    worldUpAxis = mx.Vector(upAxis)

    return [aimAxis, upAxis, worldUpAxis]
