# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""Utility functionality to working with UI."""

import os

import PySide2.QtWidgets as qtw
import PySide2.QtGui as qtg
from shiboken2 import wrapInstance

import maya.OpenMayaUI as oldomui


def getMayaMainWindow():
    """Get the Maya main window as `QMainWindow`."""
    mainWindowPtr = oldomui.MQtUtil.mainWindow()
    return wrapInstance(int(mainWindowPtr), qtw.QMainWindow)


def findWidgetByName(widgetName):
    """Find a widget by its `objectName()`."""
    mayaWindow = getMayaMainWindow()
    return mayaWindow.findChild(qtw.QWidget, widgetName)


def findWidgetsByClass(cls):
    """Find a list of widgets of type `cls`."""
    mayaWindow = getMayaMainWindow()
    return mayaWindow.findChildren(cls)


def addWidgetToMayaLayout(widgetName, parentName, keepOriginalName=False):
    """Add a widget to an existing Maya layout.

    Maya may modify the widget's name to ensure uniqueness.
    To avoid this set `keepOriginalName` to ``True``.
    """
    widgetPtr = oldomui.MQtUtil.findControl(widgetName)
    parentPtr = oldomui.MQtUtil.findControl(parentName)
    oldomui.MQtUtil.addWidgetToMayaLayout(int(widgetPtr), int(parentPtr))

    if keepOriginalName:
        # MQtUtil.addWidgetToMayaLayout may modify the objectName
        # to ensure uniqueness.
        widget = wrapInstance(int(widgetPtr), qtw.QWidget)
        widget.setObjectName(widgetName)


def clearLayout(layout):
    """Remove all content from a `QLayout`."""
    while layout.count():
        child = layout.takeAt(0)

        if child.widget():
            child.widget().deleteLater()
        elif child.layout():
            clearLayout(child.layout())


def loadIcon(name):
    """Load an icon from the default icons directory."""
    if name.startswith(':'):
        iconPath = name
    else:
        filename = os.path.join(os.path.dirname(__file__), '..', '..', '..', 'icons', name)
        extension = '.svg' if not name.endswith(('.svg', '.png')) else ''
        iconPath = filename + extension

    return qtg.QIcon(iconPath)
