# --------------------------------------------------------------------------- #
# Copyright (c) 2021 MOVINDOT ART S.R.L. (movindot.com). All rights reserved. #
# See eula.txt in the project root for license information.                   #
# --------------------------------------------------------------------------- #

"""Custom UI windows."""

import os
import re
import PySide2.QtCore as qt
import PySide2.QtWidgets as qtw
import maya.cmds as cmds
import puppeteer.utils as utils
from puppeteer.ui.utils import getMayaMainWindow, findWidgetsByClass, addWidgetToMayaLayout
from . import qssx
from .widgets import SCALE_FACTOR


class Window(qtw.QWidget):
    """The base class for all the windows."""

    def __init__(self, title):
        super().__init__(parent=getMayaMainWindow())

        self.setWindowFlags(qt.Qt.Window)
        self.setWindowTitle(title)

        # TODO: Remove if the UI scaling gets fixed in Maya > 2022.1.
        toolbarIconSize = int(qtw.QToolBar().iconSize().width() * SCALE_FACTOR)
        buttonIconSize = int(qtw.QPushButton().iconSize().width() * SCALE_FACTOR)
        checkboxSize = int(12 * SCALE_FACTOR)
        self.setStyleSheet(
            f'QToolBar {{ qproperty-iconSize: {toolbarIconSize}px }}'
            f'QPushButton {{ qproperty-iconSize: {buttonIconSize}px }}'
            f'QCheckBox::indicator {{ width: {checkboxSize}px; height: {checkboxSize}px }}'
        )

        self.addStyleSheet('base')

    def addStyleSheet(self, baseName):
        """Load a new style sheet on top of the existing styles.

        The style sheet file will be loaded from the default styles directory.

        Parameters
        ----------
        baseName : str
            The file base name, without extension.

        Raises
        ------
        IOError
            If the style sheet file could not be found.
        """
        filePath = os.path.join(os.path.dirname(__file__), '..', '..', 'qss', f'{baseName}.qss')
        fileDirPath = os.path.dirname(filePath).replace('\\', '/')

        try:
            with open(filePath) as styleFile:
                styleData = re.sub(
                    r"url\('([^:].*)'\)",
                    rf"url('{fileDirPath}/\1')",
                    styleFile.read(),
                )

                styleData = qssx.parse(styleData, {'scaleFactor': SCALE_FACTOR})

                self.setStyleSheet(self.styleSheet() + styleData)
        except IOError:
            raise IOError(f'The "{baseName}" style sheet file does not exist.')

    def closeEvent(self, event):
        """Call when Qt receives a window close request."""
        super().closeEvent(event)

        self.onClose()

    def onClose(self):
        """Call when the window gets closed."""


# --------------------------------------------------------------------------------------------------


class _SingletonWindowMetaclass(type(Window)):
    def __call__(cls, *args, **kwargs):
        try:
            return findWidgetsByClass(cls)[0]
        except IndexError:
            return super().__call__(*args, **kwargs)


class SingletonWindow(Window, metaclass=_SingletonWindowMetaclass):
    """Create a window that can have only one instance opened."""

    def __init__(self, title):
        super().__init__(title)
        self.setObjectName(utils.getClassFullName(self.__class__))


# --------------------------------------------------------------------------------------------------


class DockableWindow(SingletonWindow):
    """Create a window that can be docked."""

    def __init__(self, title, initialWidth=300, initialHeight=300):
        super().__init__(title)

        self._initialWidth = initialWidth
        self._initialHeight = initialHeight

        self._workspaceName = f'{self.objectName()}_workspaceControl'

    def createWorkspace(self):
        """Create a `maya.cmds.workspaceControl` to dock the window."""
        cls = self.__class__

        uiScript = f'from {cls.__module__} import {cls.__name__};{cls.__name__}().show()'

        closeCommand = f'from {cls.__module__} import {cls.__name__};{cls.__name__}().onClose()'

        cmds.workspaceControl(
            self._workspaceName,
            label=self.windowTitle(),
            retain=False,
            loadImmediately=False,
            restore=True,
            uiScript=uiScript,
            closeCommand=closeCommand,
            initialWidth=self._initialWidth,
            initialHeight=self._initialHeight,
        )

    def editWorkspace(self, **kwargs):
        """Modify the `workspaceControl`."""
        cmds.workspaceControl(self._workspaceName, edit=True, **kwargs)

    def queryWorkspace(self, **kwargs):
        """Retrieve the properties of the `workspaceControl`."""
        return cmds.workspaceControl(self._workspaceName, query=True, **kwargs)

    def show(self):
        """Show the window."""
        if self.queryWorkspace(exists=True):
            self.editWorkspace(restore=True)
        else:
            self.createWorkspace()

        addWidgetToMayaLayout(self.objectName(), self._workspaceName, True)

    def hide(self):
        """Hide the window."""
        if self.queryWorkspace(exists=True):
            self.editWorkspace(visible=False)

    def close(self):
        """Close and destroy the window."""
        if self.queryWorkspace(exists=True):
            self.editWorkspace(close=True)
            return True

        return super().close()
